with Ada.Unchecked_Deallocation;
with Erl_Interface.Erl_Eterm;    use Erl_Interface.Erl_Eterm;
with Interfaces;                 use Interfaces;

package Erl_Interface.Erl_Connect is

   Erl_Timeout : exception;

   type In_Addr is record
      B1, B2, B3, B4 : Unsigned_8;
   end record;

   type File_Descriptor is new Integer;

   Erl_Tick : constant Integer;
   Erl_Msg  : constant Integer;

   type Message_Kind is
     (Erl_Send, Erl_Reg_Send, Erl_Link, Erl_Unlink, Erl_Exit);

   type String_Access is access String;

   procedure Free is new Ada.Unchecked_Deallocation (String, String_Access);

   type Emessage is record
      Kind    : Message_Kind;
      Msg     : Eterm_Ptr;
      To      : Eterm_Ptr;
      From    : Eterm_Ptr;
      To_Name : String_Access;
   end record;

   procedure Erl_Connect_Init
     (Number   : in Integer;
      Cookie   : in String;
      Creation : in Integer);
   procedure Erl_Connect_Xinit
     (Host     : in String;
      Alive    : in String;
      Node     : in String;
      Addr     : in In_Addr;
      Cookie   : in String;
      Creation : in Integer);
   --  Prepare connection to a node, or raise Erl_Error if a failure occurs

   function Erl_Connect (Node : String) return File_Descriptor;
   function Erl_Xconnect
     (Addr : In_Addr; Alive : String)
     return File_Descriptor;
   --  Connect to a node, or raise Erl_Error

   procedure Erl_Close_Connection (FD : in out File_Descriptor);
   --  Close an Erlang connection. Can raise Erl_Error.

   function Erl_Receive
     (FD     : File_Descriptor;
      Buffer : access Buffer_Type)
     return Natural;
   --  Return the number of bytes read, or raise Erl_Error. If the number is
   --  Erl_Tick, then a tick has been read instead.

   function Erl_Receive_Msg
     (FD      : File_Descriptor;
      Buffer  : access Buffer_Type;
      Message : access Emessage)
     return Integer;
   --  Get a message from the queue and return Erl_Tick or Erl_Msg. Can
   --  also raise Erl_Error.

   --  erl_receive_xmsg not handled, needs a buffer type with a special
   --  pool ???

   procedure Erl_Send
     (FD  : in File_Descriptor;
      To  : access Eterm;
      Msg : access Eterm);
   procedure Erl_Reg_Send
     (FD  : in File_Descriptor;
      To  : in String;
      Msg : access Eterm);
   --  Send a message. Can raise Erl_Error.

   function Erl_RPC
     (FD     : File_Descriptor;
      Module : String;
      Func   : String;
      Args   : access Eterm)
     return Eterm_Ptr;
   --  Make a RPC

   procedure Erl_RPC_To
     (FD     : File_Descriptor;
      Module : String;
      Func   : String;
      Args   : access Eterm);
   --  Initiate a RPC

   procedure Erl_RPC_From
     (FD      : in File_Descriptor;
      Timeout : in Duration;
      Emsg    : access Emessage);
   --  Get the result of a RPC. Can raise Erl_Error or Erl_Timeout.

   function Erl_Publish (Port : Unsigned_16) return File_Descriptor;
   --  Publish this node to EPMD and get a descriptor back

private

   Erl_Tick : constant Integer := 0;
   Erl_Msg  : constant Integer := 1;  --  To check

   pragma Convention (C, In_Addr);
   for In_Addr'Size use 32;

end Erl_Interface.Erl_Connect;
