with Ada.Unchecked_Conversion;
with Interfaces.C.Strings;     use Interfaces.C, Interfaces.C.Strings;
with System;                   use System;

package body Erl_Interface.Erl_Eterm is

   function ada_erl_int_value (Term : access Eterm) return int;
   pragma Import (C, ada_erl_int_value, "ada_erl_int_value");

   function ada_erl_int_uvalue (Term : access Eterm) return unsigned;
   pragma Import (C, ada_erl_int_uvalue, "ada_erl_int_uvalue");

   function ada_erl_bin_size (Term : access Eterm) return int;
   pragma Import (C, ada_erl_bin_size, "ada_erl_bin_size");

   ----------
   -- Cons --
   ----------

   function Cons (Head : Eterm_Ptr; Tail : Eterm_Ptr) return Eterm_Ptr is
      function erl_cons (Head : Eterm_Ptr; Tail : Eterm_Ptr) return Eterm_Ptr;
      pragma Import (C, erl_cons, "erl_cons");
   begin
      case Kind (Tail) is

         when Eterm_Empty_List | Eterm_Cons =>
            return erl_cons (Head, Tail);

         when others =>
            raise Erl_Error;

      end case;
   end Cons;

   --------------
   -- Creation --
   --------------

   function Creation (Term : access Eterm) return Integer is
      function ada_erl_pid_creation (Term : access Eterm) return int;
      pragma Import (C, ada_erl_pid_creation, "ada_erl_pid_creation");
      function ada_erl_port_creation (Term : access Eterm) return int;
      pragma Import (C, ada_erl_port_creation, "ada_erl_port_creation");
   begin
      case Kind (Term) is

         when Eterm_PID =>
            return Integer (ada_erl_pid_creation (Term));

         when Eterm_Port =>
            return Integer (ada_erl_port_creation (Term));

         when others =>
            raise Erl_Error;

      end case;
   end Creation;

   -------------
   -- Element --
   -------------

   function Element (Term : access Eterm; Nth : Positive) return Eterm_Ptr
   is
      function erl_element (Position : int; Tuple : access Eterm)
        return Eterm_Ptr;
      pragma Import (C, erl_element, "erl_element");
   begin
      if Kind (Term) = Eterm_Tuple and then Size (Term) >= Nth then
         return erl_element (int (Nth), Term);
      else
         raise Erl_Error;
      end if;
   end Element;

   ----------
   -- Free --
   ----------

   procedure Free (Term : in out Eterm_Ptr) is
      procedure erl_free (Term : in Eterm_Ptr);
      pragma Import (C, erl_free, "erl_free");
   begin
      erl_free (Term);
      Term := null;
   end Free;

   -------------------
   -- Free_Compound --
   -------------------

   procedure Free_Compound (Term : in out Eterm_Ptr) is
      procedure erl_free_compound (Term : in Eterm_Ptr);
      pragma Import (C, erl_free_compound, "erl_free_compound");
   begin
      erl_free_compound (Term);
      Term := null;
   end Free_Compound;

   ---------
   -- Get --
   ---------

   function Get (Term : access Eterm) return String is
      function ada_erl_atom_ptr (Term : access Eterm) return chars_ptr;
      pragma Import (C, ada_erl_atom_ptr, "ada_erl_atom_ptr");
      function erl_iolist_to_string
        (Term : access Eterm)
        return chars_ptr;
      pragma Import (C, erl_iolist_to_string, "erl_iolist_to_string");
      procedure erl_free (Str : in chars_ptr);
      pragma Import (C, erl_free, "erl_free");
   begin
      case Kind (Term) is

         when Eterm_Atom =>
            return Value (ada_erl_atom_ptr (Term));

         when Eterm_Cons =>
            declare
               C_Res : constant chars_ptr := erl_iolist_to_string (Term);
               Res   : constant String    := Value (C_Res);
            begin
               erl_free (C_Res);
               return Res;
            end;

         when Eterm_Empty_List =>
            return "";

         when others =>
            raise Erl_Error;

      end case;
   end Get;

   ---------
   -- Get --
   ---------

   function Get (Term : access Eterm) return Integer is
   begin
      case Kind (Term) is

        when Eterm_Integer =>
          return Integer (ada_erl_int_value (Term));

         when Eterm_Unsigned_Integer =>
            return Integer (ada_erl_int_uvalue (Term));

         when others =>
            raise Erl_Error;

      end case;
   end Get;

   ---------
   -- Get --
   ---------

   function Get (Term : access Eterm) return Unsigned_32 is
   begin
      case Kind (Term) is

         when Eterm_Unsigned_Integer =>
            return Unsigned_32 (ada_erl_int_uvalue (Term));

        when Eterm_Integer =>
          return Unsigned_32 (ada_erl_int_value (Term));

         when others =>
            raise Erl_Error;

      end case;
   end Get;

   ---------
   -- Get --
   ---------

   function Get (Term : access Eterm) return Float is
      function ada_erl_float_value (Term : access Eterm) return double;
      pragma Import (C, ada_erl_float_value, "ada_erl_float_value");
   begin
      return Float (ada_erl_float_value (Term));
   end Get;

   ---------
   -- Get --
   ---------

   function Get (Term : access Eterm) return Buffer_Type is
      function ada_erl_bin_ptr (Term : access Eterm) return Address;
      pragma Import (C, ada_erl_bin_ptr, "ada_erl_bin_ptr");
      Size          : constant Natural := Natural (ada_erl_bin_size (Term));
      C_Result_Addr : constant Address := ada_erl_bin_ptr (Term);
      C_Result      : Buffer_Type (1 .. Size);
      for C_Result'Address use C_Result_Addr;
   begin
      return C_Result;
   end Get;

   ----------
   -- Head --
   ----------

   function Head (Term : access Eterm) return Eterm_Ptr is
      function erl_hd (Term : access Eterm) return Eterm_Ptr;
      pragma Import (C, erl_hd, "erl_hd");
   begin
      case Kind (Term) is

         when Eterm_Cons =>
           return erl_hd (Term);

         when others =>
            raise Erl_Error;

      end case;
   end Head;

   ----------
   -- Kind --
   ----------

   function Kind (Term : access Eterm) return Eterm_Kind is
      function ada_erl_kind (Term : access Eterm) return int;
      pragma Import (C, ada_erl_kind, "ada_erl_kind");
   begin
      return Eterm_Kind'Val (ada_erl_kind (Term));
   end Kind;

   ------------
   -- Length --
   ------------

   function Length (Term : access Eterm) return Natural is
      function erl_length (Term : access Eterm) return int;
      pragma Import (C, erl_length, "erl_length");
   begin
      case Kind (Term) is

        when Eterm_Cons =>
          return Natural (erl_length (Term));

         when Eterm_Empty_List =>
            return 0;

         when others =>
            raise Erl_Error;

      end case;
   end Length;

   ---------------
   -- Make_Atom --
   ---------------

   function Make_Atom (Atom : String) return Eterm_Ptr is
      function erl_mk_atom (Atom : Address) return Eterm_Ptr;
      pragma Import (C, erl_mk_atom, "erl_mk_atom");
   begin
      return erl_mk_atom (To_C (Atom)'Address);
   end Make_Atom;

   -----------------
   -- Make_Binary --
   -----------------

   function Make_Binary (Buffer : Buffer_Type) return Eterm_Ptr is
      function erl_mk_binary (Addr : Address; Size : int) return Eterm_Ptr;
      pragma Import (C, erl_mk_binary, "erl_mk_binary");
   begin
      return erl_mk_binary (Buffer'Address, Buffer'Length);
   end Make_Binary;

   ----------------
   -- Make_Float --
   ----------------

   function Make_Float (From : Float) return Eterm_Ptr is
      function erl_mk_float (From : double) return Eterm_Ptr;
      pragma Import (C, erl_mk_float, "erl_mk_float");
   begin
      return erl_mk_float (double (From));
   end Make_Float;

   ------------------
   -- Make_Integer --
   ------------------

   function Make_Integer (From : Integer) return Eterm_Ptr is
      function erl_mk_int (From : int) return Eterm_Ptr;
      pragma Import (C, erl_mk_int, "erl_mk_int");
   begin
      return erl_mk_int (int (From));
   end Make_Integer;

   ---------------
   -- Make_List --
   ---------------

   function Make_List (From : Eterm_Ptr_Array) return Eterm_Ptr is
      function erl_mk_list (Addr : Address; Length : int) return Eterm_Ptr;
      pragma Import (C, erl_mk_list, "erl_mk_list");
   begin
      return erl_mk_list (From'Address, From'Length);
   end Make_List;

   --------------
   -- Make_PID --
   --------------

   function Make_PID
     (Node : String; Number : Integer; Serial : Integer; Creation : Integer)
     return Eterm_Ptr
   is
      function erl_mk_pid
        (Node : Address; Number : int; Serial : int; Creation : int)
        return Eterm_Ptr;
      pragma Import (C, erl_mk_pid, "erl_mk_pid");
   begin
      return erl_mk_pid (To_C (Node)'Address, int (Number), int (Serial),
                         int (Creation));
   end Make_PID;

   ---------------
   -- Make_Port --
   ---------------

   function Make_Port (Node : String; Number : Integer; Creation : Integer)
     return Eterm_Ptr
   is
      function erl_mk_port
        (Node : Address; Number : int; Creation : int)
        return Eterm_Ptr;
      pragma Import (C, erl_mk_port, "erl_mk_port");
   begin
      return erl_mk_port (To_C (Node)'Address, int (Number), int (Creation));
   end Make_Port;

   -----------------
   -- Make_String --
   -----------------

   function Make_String (From : String) return Eterm_Ptr is
      function erl_mk_estring (Addr : Address; Size : int) return Eterm_Ptr;
      pragma Import (C, erl_mk_estring, "erl_mk_estring");
   begin
      return erl_mk_estring (From'Address, From'Length);
   end Make_String;

   ----------------
   -- Make_Tuple --
   ----------------

   function Make_Tuple (From : Eterm_Ptr_Array) return Eterm_Ptr is
      function erl_mk_tuple (Addr : Address; Length : int) return Eterm_Ptr;
      pragma Import (C, erl_mk_tuple, "erl_mk_tuple");
   begin
      return erl_mk_tuple (From'Address, From'Length);
   end Make_Tuple;

   ---------------------------
   -- Make_Unsigned_Integer --
   ---------------------------

   function Make_Unsigned_Integer (From : Unsigned_32) return Eterm_Ptr is
      function erl_mk_uint (From : unsigned) return Eterm_Ptr;
      pragma Import (C, erl_mk_uint, "erl_mk_uint");
   begin
      return erl_mk_uint (unsigned (From));
   end Make_Unsigned_Integer;

   --------------
   -- Make_Var --
   --------------

   function Make_Var (Name : String) return Eterm_Ptr is
      function erl_mk_var (Name : Address) return Eterm_Ptr;
      pragma Import (C, erl_mk_var, "erl_mk_var");
   begin
      return erl_mk_var (To_C (Name)'Address);
   end Make_Var;

   ----------
   -- Node --
   ----------

   function Node (Term : access Eterm) return Eterm_Ptr is
      function ada_erl_pid_node (Term : access Eterm) return Eterm_Ptr;
      pragma Import (C, ada_erl_pid_node, "ada_erl_pid_node");
      function ada_erl_port_node (Term : access Eterm) return Eterm_Ptr;
      pragma Import (C, ada_erl_port_node, "ada_erl_port_node");
   begin
      case Kind (Term) is

         when Eterm_PID =>
            return ada_erl_pid_node (Term);

         when Eterm_Port =>
            return ada_erl_port_node (Term);

         when others =>
            raise Erl_Error;

      end case;
   end Node;

   ------------
   -- Number --
   ------------

   function Number (Term : access Eterm) return Integer is
      function ada_erl_pid_number (Term : access Eterm) return int;
      pragma Import (C, ada_erl_pid_number, "ada_erl_pid_number");
      function ada_erl_port_number (Term : access Eterm) return int;
      pragma Import (C, ada_erl_port_number, "ada_erl_port_number");
   begin
      case Kind (Term) is

         when Eterm_PID =>
            return Integer (ada_erl_pid_number (Term));

         when Eterm_Port =>
            return Integer (ada_erl_port_number (Term));

         when others =>
            raise Erl_Error;

      end case;
   end Number;

   ------------
   -- Serial --
   ------------

   function Serial (Term : access Eterm) return Integer is
      function ada_erl_pid_serial (Term : access Eterm) return int;
      pragma Import (C, ada_erl_pid_serial, "ada_erl_pid_serial");
   begin
      case Kind (Term) is

         when Eterm_PID =>
            return Integer (ada_erl_pid_serial (Term));

         when others =>
            raise Erl_Error;

      end case;
   end Serial;

   ----------
   -- Size --
   ----------

   function Size (Term : access Eterm) return Natural is
      function ada_erl_tuple_size (Term : access Eterm) return int;
      pragma Import (C, ada_erl_tuple_size, "ada_erl_tuple_size");
   begin
      case Kind (Term) is

        when Eterm_Tuple =>
          return Natural (ada_erl_tuple_size (Term));

         when Eterm_Binary =>
            return Natural (ada_erl_bin_size (Term));

         when others =>
            raise Erl_Error;

      end case;
   end Size;

   ----------
   -- Tail --
   ----------

   function Tail (Term : access Eterm) return Eterm_Ptr is
      function erl_tl (Term : access Eterm) return Eterm_Ptr;
      pragma Import (C, erl_tl, "erl_tl");
   begin
      case Kind (Term) is

         when Eterm_Cons =>
           return erl_tl (Term);

         when others =>
            raise Erl_Error;

      end case;
   end Tail;

   -----------------
   -- Var_Content --
   -----------------

   function Var_Content (Term : access Eterm; Name : String) return Eterm_Ptr
   is
      function erl_var_content (Term : access Eterm; Name : Address)
        return Eterm_Ptr;
      pragma Import (C, erl_var_content, "erl_var_content");
      Result : constant Eterm_Ptr :=
        erl_var_content (Term, To_C (Name)'Address);
   begin
      if Result = null then
         raise Erl_Error;
      else
         return Result;
      end if;
   end Var_Content;

   -----------
   -- Write --
   -----------

   procedure Write
     (Stream : access Root_Stream_Type'Class;
      Term   : in Eterm_Ptr)
   is
      function To_Int is
         new Ada.Unchecked_Conversion (Eterm_Ptr, int);
   begin
      int'Write (Stream, To_Int (Term));
   end Write;

   procedure ada_erl_init;
   pragma Import (C, ada_erl_init, "ada_erl_init");
begin
   ada_erl_init;
end Erl_Interface.Erl_Eterm;
