with Ada.Streams; use Ada.Streams;
with Interfaces;  use Interfaces;

package Erl_Interface.Erl_Eterm is

   type Eterm is limited private;

   type Eterm_Ptr is access Eterm;
   pragma Convention (C, Eterm_Ptr);
   procedure Read
     (Stream : access Root_Stream_Type'Class;
      Term   : out Eterm_Ptr);
   pragma Import (Stubbed, Read);
   procedure Write
     (Stream : access Root_Stream_Type'Class;
      Term   : in Eterm_Ptr);
   for Eterm_Ptr'Read use Read;
   for Eterm_Ptr'Write use Write;
   --  Those pointers must be storable in a stream, for formatting purpose.
   --  Note that the Read function will raise (on purpose) Program_Error,
   --  as this type cannot be stored in the stream in spite of the presence
   --  of those two procedures.

   procedure Free (Term : in out Eterm_Ptr);
   --  Free an Erlang term

   procedure Free_Compound (Term : in out Eterm_Ptr);
   --  Recursively free an Erlang term

   type Eterm_Ptr_Array is array (Natural range <>) of Eterm_Ptr;
   pragma Convention (C, Eterm_Ptr_Array);

   type Eterm_Kind is
     (Eterm_Integer, Eterm_Unsigned_Integer, Eterm_Float, Eterm_Atom,
      Eterm_PID, Eterm_Port, Eterm_Ref, Eterm_Tuple, Eterm_Binary,
      Eterm_Empty_List, Eterm_Cons, Eterm_Unknown);
   --  Should the list change, ada_erl_glue.c must be adapted as well

   type Buffer_Type is array (Natural range <>) of Unsigned_8;
   pragma Pack (Buffer_Type);
   pragma Convention (C, Buffer_Type);

   function Kind (Term : access Eterm) return Eterm_Kind;
   --  Return the kind of Eterm

   --  Note: all the Get function can raise Erl_Error if the term is not
   --  appropriate

   function Get (Term : access Eterm) return String;
   --  Return a string from an Atom or a list of characters

   function Get (Term : access Eterm) return Integer;
   --  Return an integer

   function Get (Term : access Eterm) return Unsigned_32;
   --  Return an unsigned integer

   function Get (Term : access Eterm) return Float;
   --  Return a float

   function Get (Term : access Eterm) return Buffer_Type;
   --  Return a buffer from a binary

   function Size (Term : access Eterm) return Natural;
   --  Return the size of a binary or a tuple

   function Length (Term : access Eterm) return Natural;
   --  Return the length of a list

   function Head (Term : access Eterm) return Eterm_Ptr;
   function Tail (Term : access Eterm) return Eterm_Ptr;
   --  Return the head and the tail of a cons

   function Node (Term : access Eterm) return Eterm_Ptr;
   function Number (Term : access Eterm) return Integer;
   function Serial (Term : access Eterm) return Integer;
   function Creation (Term : access Eterm) return Integer;
   --  Get the node, number or creation of a PID or a port, or the serial of
   --  a PID.

   function Cons (Head : Eterm_Ptr; Tail : Eterm_Ptr) return Eterm_Ptr;
   --  Build a list from its head and its tail

   function Copy (Term : access Eterm) return Eterm_Ptr;
   --  Copy a term

   function Element (Term : access Eterm; Nth : Positive) return Eterm_Ptr;
   --  Return the nth element of a tuple

   function Var_Content (Term : access Eterm; Name : String) return Eterm_Ptr;
   --  Return the variable of name Name contained in term Term. Can raise
   --  Erl_Error.

   function Make_Atom (Atom : String) return Eterm_Ptr;
   function Make_Binary (Buffer : Buffer_Type) return Eterm_Ptr;
   function Make_Empty_List return Eterm_Ptr;
   function Make_String (From : String) return Eterm_Ptr;
   function Make_Float (From : Float) return Eterm_Ptr;
   function Make_Integer (From : Integer) return Eterm_Ptr;
   function Make_List (From : Eterm_Ptr_Array) return Eterm_Ptr;
   function Make_PID
     (Node : String; Number : Integer; Serial : Integer; Creation : Integer)
     return Eterm_Ptr;
   function Make_Port (Node : String; Number : Integer; Creation : Integer)
     return Eterm_Ptr;
   function Make_Tuple (From : Eterm_Ptr_Array) return Eterm_Ptr;
   function Make_Unsigned_Integer (From : Unsigned_32) return Eterm_Ptr;
   function Make_Var (Name : String) return Eterm_Ptr;

private

   type Eterm is limited null record;

   pragma Import (C, Copy, "erl_copy_term");
   pragma Import (C, Make_Empty_List, "erl_mk_empty_list");

   pragma Linker_Options ("ada_erl_glue.o");

end Erl_Interface.Erl_Eterm;
